/* Needham-Schroeder protocol */

/* message types */
mtype = {m1, m2, m3};

/* public keys */
mtype = {kA, kB, kI};

/* private nounces */
mtype = {nA, nB, nI, none};

typedef Crypt
{ /* the encrypted part of a message */
  mtype d0, d1, key;
}

/* Message types vary in practice according to the state of the
   session:
     <m1, <kA, nA,   kB>>
     <m2, <nA, nB,   kA>>
     <m3, <nB, none, kB>>
*/

/* communication via rendez-vous */
chan network = [0] of {mtype, /* message type in {m1,m2,m3} */
                       Crypt  /* encrypted part */};

/* state of Alice at the end of a session */
bit doneA;
  /* partnerA is the partner Alice believes she is communicating with, and
     nounceA  is the nounce Alice received from her partner */
mtype partnerA, nounceA;
/* one session of a honest principal of type `Alice' */
proctype Alice ()
{
  /* nondeterministically choose a partner */
  if
  :: skip -> partnerA = kI;
  :: skip -> partnerA = kB;
  fi

  /* send a message of type 1 to chosen partner */
  network ! m1(kA, nA, partnerA);

  /* receive message 2 */
  network ? m2(eval(nA), nounceA, eval(kA));

  /* send message 3 */
  network ! m3(nounceA, none, partnerA);

  doneA = 1;
}

/* state of Bob at the end of a session */
bit doneB;
  /* partnerB is the partner Bob believes he is communicating with, and
     nounceB  is the nounce Bob received from his partner */
mtype partnerB, nounceB;
/* one session of a honest principal of type `Bob' */
proctype Bob ()
{
  /* partnerB is the partner Bob believes he is communicating with, and
     nounceB  is the nounce Bob received from his partner */

  /* receive message 1 */
  network ? m1(partnerB, nounceB, eval(kB));

  /* send message 2 */
  network ! m2(nounceB, nB, partnerB);

  /* receive message 3 */
  network ? m3(eval(nB), none, eval(kB));

  doneB = 1;
}

proctype Intruder ()
{
  /* Intruder only needs bounded memory */
  mtype d[3];
  byte i;
  Crypt message;

  /* initialise Intruder's memory with dummy values */
  d_step {
   d[0] = nI; d[1] = nI; d[2] = kI;
   message.d0 = nI; message.d1 = nI; message.key = kI
  };

  do
  /* intercept message */
  :: skip -> network ? _(message)
  /* decypher message if encrypted with Intruder's public key */
  :: message.key == kI -> d_step { d[0] = message.d0; d[1] = message.d1 }
  /* forge new message */
  :: skip -> d_step { message.d0 = d[0]; message.d1 = d[1]; message.key = d[2] }
  /* send message */
  :: skip ->
  atomic {
    if
    :: skip -> network ! m1(message)
    :: skip -> network ! m2(message)
    :: skip -> network ! m3(message)
    fi
  }
  /* pick a different public key */
  :: skip -> 
  atomic {
    if
    :: skip -> d[2] = kA
    :: skip -> d[2] = kB
    :: skip -> d[2] = kI
    fi
  }
  od
}

/* check for a man-in-the-middle attack */
proctype monitor()
{
  atomic {
    (doneA && doneB && partnerB == kA) -> assert(partnerA == kB)
  }
}

init
{
  atomic
    {
      doneA = 0;
      doneB = 0;
      run Alice();
      run Bob();
      run Intruder();
      run monitor();
    }
}